from pathlib import Path

import pytest

from i3configger import exc, partials

SCHEMES = Path(__file__).parents[1] / 'examples' / '4-schemes' / 'config.d'


def test_create():
    prts = partials.create(SCHEMES)
    assert len(prts) == 3


@pytest.mark.parametrize(
    "key, value, exp", (
        ('', '', exc.ConfigError),
        ('some-key', '', exc.ConfigError),
        ('', '', exc.ConfigError),
        (None, None, exc.ConfigError),
        ('non-existing-key', 'some-value', exc.ConfigError),
        ('non-existing-key', 'none-existing-value', exc.ConfigError),
        ('some-category', 'none-existing-value', exc.ConfigError),
        ('some-category', 'value1', True),
        ('some-category', 'value2', True),
    )
)
def test_select_non_existing(key, value, exp):
    prts = partials.create(SCHEMES)
    selector = {key: value}
    if not isinstance(exp, bool):
        with pytest.raises(exp):
            partials.select(prts, selector)
    else:
        selected = partials.select(prts, selector)
        found = partials.find(prts, key, value)
        assert selected[1].name == 'some-settings'
        assert isinstance(selected[0], partials.Partial)
        assert isinstance(found, partials.Partial)
        assert selected[0] == found
        assert selected[0].key == key
        assert selected[0].value == value



import logging
import pprint
import socket
import typing as t
from functools import total_ordering
from pathlib import Path

from i3configger import base, exc

log = logging.getLogger(__name__)

# TODO document this
SPECIAL_SELECTORS = {
    "hostname": socket.gethostname()
}
# TODO document this
EXCLUDE_MARKER = "."
"""config files starting with a dot are always excluded"""


@total_ordering
class Partial:
    def __init__(self, path: Path):
        self.path = path
        self.name = self.path.stem
        self.selectors = self.name.split('.')
        self.needsSelection = len(self.selectors) > 1
        self.key = self.selectors[0] if self.needsSelection else None
        self.value = self.selectors[1] if self.needsSelection else None
        self.lines = self.path.read_text().splitlines()

    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__, self.path.name)

    __str__ = __repr__

    def __lt__(self, other):
        return self.name < other.name

    @property
    def display(self) -> t.Optional[str]:
        lines = [l for l in self.lines
                 if not l.strip().startswith(base.SET_MARK)]
        if not lines:
            return
        return "### %s ###\n%s\n\n" % (self.path.name, "\n".join(lines))

    @property
    def context(self):
        ctx = {}
        for line in [l.strip() for l in self.lines
                     if l.strip().startswith(base.SET_MARK)]:
            payload = line.split(maxsplit=1)[1]
            key, value = payload.split(maxsplit=1)
            ctx[key] = value
        return ctx


def find(prts: t.List[Partial], key: str, value: str= None) \
        -> t.Union[Partial, t.List[Partial]]:
    findings = []
    for prt in prts:
        if prt.key != key:
            continue
        if prt.value == value:
            return prt
        elif not value:
            findings.append(prt)
    return findings


def select(partials, selection, excludes=None) -> t.List[Partial]:
    def _select():
        selected.append(partial)
        if partial.needsSelection:
            del selection[partial.key]

    for key, value in SPECIAL_SELECTORS.items():
        if key not in selection:
            selection[key] = value
    selected = []
    for partial in partials:
        if partial.needsSelection:
            if excludes and partial.key in excludes:
                log.debug("[IGNORE] %s (in %s)", partial, excludes)
                continue
            if (selection and partial.key in selection and
                    partial.value == selection.get(partial.key)):
                _select()
        else:
            _select()
    log.debug("selected:\n%s", pprint.pformat(selected))
    if selection and not all(k in SPECIAL_SELECTORS for k in selection):
        raise exc.ConfigError(
            "selection processed incompletely: %s", selection)
    return selected


def create(partialsPath) -> t.List[Partial]:
    partialsPath = Path(partialsPath)
    assert partialsPath.is_dir(), partialsPath
    prts = []
    for path in partialsPath.glob('*%s' % base.SUFFIX):
        if path.name.startswith(EXCLUDE_MARKER):
            log.warning("exclude %s - starts with a dot", path)
            continue
        prts.append(Partial(path))
    if not prts:
        raise exc.PartialsError(f"no '*{base.SUFFIX}' at {partialsPath}")
    return sorted(prts)



import json
from pathlib import Path

import i3configger.paths
from i3configger import paths


def test_no_config(tmpdir, monkeypatch):
    """Given empty sources directory a new config is created from defaults"""
    monkeypatch.setattr(paths, 'get_i3_config_path', lambda: Path(tmpdir))
    path = i3configger.paths.get_my_config_path()
    assert path.exists()
    assert path.is_file()
    assert path.name == i3configger.paths.MY_CONFIG_NAME
    payload = json.loads(path.read_text())
    assert 'main' in payload
    assert 'bars' in payload
    assert 'targets' in payload['bars']



import logging
import typing as t
from functools import reduce
from string import Template

from i3configger import base, exc, partials

log = logging.getLogger(__name__)


def process(item: t.Iterable[t.Union[dict, partials.Partial]]) -> dict:
    """merge contexts of an arbitrary number of items"""
    dicts = []
    for elem in item:
        dicts.append(elem if isinstance(elem, dict) else elem.context)
    return reduce(merge, dicts)


def merge(dst, src):
    """merge source into destination mapping (overwrite existing keys)"""
    for key in src:
        if key in dst:
            if isinstance(dst[key], dict) and isinstance(src[key], dict):
                merge(dst[key], src[key])
            else:
                dst[key] = src[key]
        else:
            dst[key] = src[key]
    return dst


def prune(dst, src):
    for key, value in src.items():
        if isinstance(value, dict):
            prune(dst.get(key, {}), value)
        elif key in dst:
            del dst[key]
    return dst


def resolve_variables(context: dict) -> dict:
    """If variables are set by a variable, replace them by their value."""
    if not any(v.startswith(base.VAR_MARK) for v in context.values()):
        log.debug("needs no resolution: %s", context)
        return context
    resolvedContext = {}
    seenKeys = []
    for key, value in context.items():
        if key in seenKeys:
            raise exc.DuplicateKey("%s -> %s would be overridden with %s",
                                   key, context[key], value)
        if value.startswith(base.VAR_MARK):
            try:
                resolvedContext[key] = context[value]
            except KeyError:
                log.exception("[IGNORED] %s, %s", key, value)
        else:
            resolvedContext[key] = value
    return resolvedContext


def remove_variable_markers(ctx: dict) -> dict:
    cleaned = {}
    lvm = len(base.VAR_MARK)
    for key, value in ctx.items():
        key = key[lvm:] if key.startswith(base.VAR_MARK) else key
        cleaned[key] = value
    return cleaned


def substitute(content, ctx):
    """Substitute all variables with their values.

    Works out of the box, because '$' is the standard substitution
    marker for string.Template
    """
    return Template(content).safe_substitute(ctx)



class I3configgerException(Exception):
    """Main exception with log style string formatting enhancement"""
    def __init__(self, *args, **kwargs):
        try:
            # noinspection PyArgumentList
            super().__init__(args[0] % (args[1:]), **kwargs)
        except:
            # noinspection PyArgumentList
            super().__init__(*args, **kwargs)


class BuildError(I3configgerException):
    pass


class ConfigError(I3configgerException):
    pass


class DuplicateKey(I3configgerException):
    pass


class MalformedAssignment(I3configgerException):
    pass


class MessageError(I3configgerException):
    pass


class PartialsError(I3configgerException):
    pass


class ParseError(I3configgerException):
    pass



import json
import logging
import pprint
from pathlib import Path

from i3configger import exc, paths

log = logging.getLogger(__name__)

I3_CONFIGGER_DEFAULTS = {
    "main": {
        "target": "../config",
    },
    "bars": {
        "defaults": {
            "target": "..",
            "template": "tpl",
            "key": "i3status",
            "value": "default"
        },
        "targets": {}
    }
}


class I3configgerConfig:
    def __init__(self, configPath: Path):
        p = paths.Paths(configPath)
        self.configPath = configPath
        self.partialsPath = p.root
        self.payload = fetch(self.configPath)
        targetPath = Path(self.payload["main"]["target"]).expanduser()
        if targetPath.is_absolute():
            self.mainTargetPath = targetPath.resolve()
        else:
            self.mainTargetPath = (self.partialsPath / targetPath).resolve()

    def __str__(self):
        return "%s:\n%s" % (self.__class__.__name__,
                            pprint.pformat(vars(self)))

    def get_bar_targets(self):
        """Create a resolved copy of the bar settings."""
        barTargets = {}
        barSettings = self.payload.get("bars", {})
        if not barSettings:
            return barTargets
        defaults = barSettings.get("defaults", {})
        for name, bar in barSettings["targets"].items():
            newBar = dict(bar)
            barTargets[name] = newBar
            for defaultKey, defaultValue in defaults.items():
                if defaultKey not in newBar:
                    newBar[defaultKey] = defaultValue
            container = Path(newBar["target"])
            if not container.is_absolute():
                container = (self.partialsPath / container).resolve()
            newBar["target"] = str(container)
        return barTargets


def fetch(path):
    if not path.exists():
        raise exc.ConfigError(f"file not found: {path}")
    with path.open() as f:
        log.info("fetch from %s", )
        payload = json.load(f)
        log.debug(f"{path}: {pprint.pformat(payload)}")
        return payload


def freeze(path, obj):
    with open(path, 'w') as f:
        json.dump(obj, f, sort_keys=True, indent=2)
    log.debug("froze %s to %s", pprint.pformat(obj), path)



import logging
from pathlib import Path

from i3configger import base, config, exc, partials, context

log = logging.getLogger(__name__)


I3STATUS = "i3status"
"""reserved key for status bar setting files"""
# TODO document deletion for set and shadow
DEL = 'del'
"""signal to delete a key in shadow or set"""


class CMD:
    SELECT_NEXT = "select-next"
    SELECT_PREVIOUS = "select-previous"
    SELECT = "select"
    SET = "set"
    MERGE = "merge"
    PRUNE = "prune"
    SHADOW = "shadow"


def process(statePath, prts, message):
    mp = Messenger(statePath, prts, message)
    mp.execute()
    config.freeze(statePath, mp.payload)


class Messenger:
    def __init__(self, messagesPath, prts, message=None):
        self.messagesPath = messagesPath
        self.prts = prts
        self.message = message
        self.payload = self.fetch_frozen_messages()
        if self.message:
            self.command, self.key, *rest = message
            self.value = rest[0] if rest else None
        log.debug(f"sending message {message} to {messagesPath}")

    def execute(self):
        {
            CMD.MERGE: self._process_merge,
            CMD.PRUNE: self._process_prune,
            CMD.SET: self._process_set,
            CMD.SELECT: self._process_select,
            CMD.SHADOW: self._process_shadow,
            CMD.SELECT_NEXT: self._process_select_shift,
            CMD.SELECT_PREVIOUS: self._process_select_shift,
        }[self.command]()

    def _process_merge(self):
        self._transform(context.merge)

    def _process_prune(self):
        self._transform(context.prune)

    def _transform(self, func):
        path = Path(self.key).expanduser()
        if not path.is_absolute():
            path = self.messagesPath.parent / path
        self.payload = func(self.payload, config.fetch(path))
        config.freeze(self.messagesPath, self.payload)

    def _process_set(self):
        if self.value and self.value.lower() == DEL:
            del self.payload[CMD.SET][base.VAR_MARK + self.key]
        else:
            self.payload[CMD.SET][base.VAR_MARK + self.key] = self.value

    def _process_select(self):
        candidates = partials.find(self.prts, self.key)
        if not candidates:
            raise exc.MessageError(
                f"No candidates for {self.message} in {self.prts}")
        candidate = partials.find(self.prts, self.key, self.value)
        if not candidate:
            raise exc.MessageError(
                f"No candidates for {self.message} in {candidates}")
        if self.value and self.value.lower() == DEL:
            del self.payload[CMD.SELECT][self.key]
        else:
            self.payload[CMD.SELECT][self.key] = candidate.value

    def _process_shadow(self):
        """Shadow arbitrary settings made in i3configger.json.

        key:deeper:deepest[...] -> [key][deeper][deepest][...]
        """
        parts = self.key.split(':')
        current = self.payload[CMD.SHADOW]
        while True:
            part = parts.pop(0)
            if parts:
                current[part] = {}
                current = current[part]
            else:
                if self.value is not None and self.value.lower() == DEL:
                    del current[part]
                else:
                    current[part] = self.value
                break

    def _process_select_shift(self):
        candidates = partials.find(self.prts, self.key)
        if not candidates:
            raise exc.MessageError(
                f"No candidates for {self.message} in {self.prts}")
        if self.command == CMD.SELECT_PREVIOUS:
            candidates = reversed(candidates)
        current = self.payload["select"].get(self.key) or candidates[0].key
        for idx, candidate in enumerate(candidates):
            if candidate.value == current:
                try:
                    new = candidates[idx + 1]
                except IndexError:
                    new = candidates[0]
                log.info("select %s.%s", self.key, new)
                self.payload[CMD.SELECT][self.key] = new.value
                break

    def fetch_frozen_messages(self):
        if not self.messagesPath.exists():
            state = {}
        else:
            state = config.fetch(self.messagesPath)
        self.ensure_message_keys(state, self.prts)
        config.freeze(self.messagesPath, state)
        return state

    def ensure_message_keys(self, state, prts):
        if CMD.SELECT not in state:
            initialSelects = {}
            for prt in prts:
                if not prt.needsSelection:
                    continue
                if prt.key not in initialSelects and prt.key != I3STATUS:
                    initialSelects[prt.key] = prt.value
            state[CMD.SELECT] = initialSelects
        if CMD.SET not in state:
            state[CMD.SET] = {}
        if CMD.SHADOW not in state:
            state[CMD.SHADOW] = {}



import os
import shutil
from pathlib import Path

import pytest

from i3configger import build, paths, message, partials

HERE = Path(__file__).parent
EXAMPLES = HERE.parent / 'examples'
REFERENCE = HERE / 'examples'
FAKE_HEADER = """\
###############################################################################
# Built by i3configger /from/some/directory/who/cares  (some time after 1972) #
###############################################################################
"""
TEST_FOLDER_NAMES = sorted(list(
    [d.name for d in EXAMPLES.iterdir()
     if d.is_dir() and not str(d.name).startswith('_')]))


@pytest.mark.parametrize("container", TEST_FOLDER_NAMES)
def test_build(container, monkeypatch):
    monkeypatch.setattr(
        paths, 'get_i3_config_path', lambda: EXAMPLES / container)
    monkeypatch.setattr(build, 'make_header', lambda _: FAKE_HEADER)
    if not shutil.which('i3'):
        # we're on CI
        monkeypatch.setattr(build, 'check_config', lambda _: True)
    configPath = paths.get_my_config_path()
    assert configPath.exists() and configPath.is_file()
    p = paths.Paths(configPath)
    prts = partials.create(p.root)
    message.Messenger(p.messages, prts)
    try:
        build.build_all(configPath)
        buildPath = configPath.parents[1]
        referencePath = REFERENCE / container
        names = [p.name for p in referencePath.iterdir()]
        assert names
        for name in names:
            resultFilePath = buildPath / name
            referenceFilePath = (referencePath / name)
            assert resultFilePath != referenceFilePath
            result = resultFilePath.read_text()
            reference = referenceFilePath.read_text()
            assert result == reference
    finally:
        if p.messages.exists():
            os.unlink(p.messages)



import re
import sys
from pathlib import Path
from setuptools import find_packages, setup


def linkify(changelog):
    url = 'https://github.com/obestwalter/i3configger'
    issueReplacement = (r'[^\[]#(\d+)', r'`#\1 <%s/issues/\1>`_' % url)
    pullReplacement = (r'[^\[]#p(\d+)', r'`#p\1 <%s/pull/\1>`_' % url)
    for pattern, replacement in [issueReplacement, pullReplacement]:
        changelog = re.sub(pattern, replacement, changelog)
    return changelog


def get_long_description():
    """Pypi doesn't like .md - I don't like .rst - let's compromise."""
    here = Path(__file__).parent
    readme = (here / 'README.md').read_text()
    changelog = (here / 'CHANGELOG.md').read_text()
    if 'upload' in sys.argv:
        import pypandoc
        readme = pypandoc.convert_text(readme, to='rst', format='md')
        changelog = pypandoc.convert_text(changelog, to='rst', format='md')
        changelog = linkify(changelog)
    return "%s\n\n%s" % (readme, changelog)


kwargs = dict(
    name='i3configger',
    author='Oliver Bestwalter',
    url='https://github.com/obestwalter/i3configger',
    description="i3 config manipulation tool",
    long_description=get_long_description(),
    use_scm_version=True,
    setup_requires=['setuptools_scm'],
    entry_points={'console_scripts': ['i3configger = i3configger.main:main']},
    install_requires=['inotify', 'psutil', 'python-daemon'],
    packages=find_packages(),
    classifiers=[
        'Development Status :: 4 - Beta',
        'Operating System :: POSIX :: Linux',
        'License :: OSI Approved :: MIT License',
        'Environment :: Console',
        'Topic :: Utilities',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
    ],
)


if __name__ == '__main__':
    setup(**kwargs)



import logging
import sys

from i3configger import (
    build, cli, daemonize, ipc, partials, paths, message, watch)

log = logging.getLogger(__name__)


def main():
    args = cli.process_command_line()
    if args.kill:
        daemonize.exorcise()
        return 0
    configPath = paths.get_my_config_path(args.config)
    if args.message:
        p = paths.Paths(configPath)
        prts = partials.create(p.root)
        message.process(p.messages, prts, args.message)
    if daemonize.get_other_i3configgers():
        if not args.message:
            sys.exit("already running. Did you mean to send a message?")
        return 0
    ipc.I3.set_msg_type(args.i3_refresh_msg)
    log.info("set i3 refresh method to %s", ipc.I3.refresh)
    ipc.Notify.set_notify_command(args.notify)
    if args.daemon:
        daemonize.daemonize(args.v, args.log, configPath)
        return 0
    if args.watch:
        try:
            watch.Watchman(configPath).watch()
        except KeyboardInterrupt:
            sys.exit("interrupted by user")
    build.build_all(configPath)
    ipc.I3.refresh()
    ipc.StatusBar.refresh()
    ipc.Notify.send('new config active')



import logging
import os
import tempfile
import time
from pathlib import Path

from i3configger import (
    __version__, base, config, context, exc, ipc, partials, message, paths)

log = logging.getLogger(__name__)


def build_all(configPath):
    p = paths.Paths(configPath)
    prts = partials.create(p.root)
    cnf = config.I3configgerConfig(configPath)
    msg = config.fetch(p.messages)
    cnf.payload = context.merge(cnf.payload, msg[message.CMD.SHADOW])
    pathContentsMap = generate_contents(
        cnf, prts, msg[message.CMD.SELECT], msg[message.CMD.SET])
    check_config(pathContentsMap[cnf.mainTargetPath])
    persist_results(pathContentsMap)


def generate_contents(cnf, prts, selectorMap, setMap):
    pathContentsMap = {}
    barTargets = cnf.get_bar_targets()
    excludes = {b["key"] for b in barTargets.values()}
    selected = partials.select(prts, selectorMap, excludes)
    ctx = context.process(selected + [setMap])
    ctx = context.resolve_variables(ctx)
    ctx = context.remove_variable_markers(ctx)
    mainContent = generate_main_content(cnf.partialsPath, selected, ctx)
    for barName, barCnf in barTargets.items():
        barCnf["id"] = barName
        eCtx = context.process([ctx, barCnf])
        mainContent += "\n%s" % generate_bar_setting(barCnf, prts, eCtx)
        statusFileContent = generate_status_file_content(
            prts, barCnf["key"], barCnf["value"], eCtx)
        if statusFileContent:
            filename = f"{barCnf['key']}.{barCnf['value']}{base.SUFFIX}"
            dst = Path(barCnf["target"]) / filename
            pathContentsMap[dst] = statusFileContent
    pathContentsMap[cnf.mainTargetPath] = mainContent.rstrip('\n') + '\n'
    return pathContentsMap


def make_header(partialsPath):
    strPath = str(partialsPath)
    parts = strPath.split(os.getenv('HOME'))
    if len(parts) > 1:
        strPath = "~" + parts[-1]
    msg = (f'# Built from {strPath} by i3configger {__version__} '
           f'({time.asctime()}) #')
    sep = "#" * len(msg)
    return "%s\n%s\n%s\n" % (sep, msg, sep)


def generate_main_content(partialsPath, selected, ctx):
    content = make_header(partialsPath)
    content += '\n'.join(prt.display for prt in selected if prt.display)
    return context.substitute(content, ctx).rstrip('\n') + '\n\n'


def generate_bar_setting(barCnf, prts, ctx):
    tpl = partials.find(prts, barCnf["key"], barCnf["template"])
    assert isinstance(tpl, partials.Partial), tpl
    return context.substitute(tpl.display, ctx).rstrip('\n')


def generate_status_file_content(prts, selectKey, selectValue, ctx):
    prt = partials.find(prts, selectKey, selectValue)
    if not prt:
        log.warning("[IGNORE] no status config named %s.%s%s",
                    selectKey, selectKey, base.SUFFIX)
        return
    assert isinstance(prt, partials.Partial), prt
    return context.substitute(prt.display, ctx).rstrip('\n') + '\n'


def check_config(content):
    tmpPath = Path(tempfile.gettempdir()) / 'i3config_check'
    tmpPath.write_text(content)
    errorReport = ipc.I3.get_config_errors(tmpPath)
    if errorReport:
        raise exc.ConfigError(
            f"config:\n{content}\n\nerrors:\n{errorReport.decode()}"
            f"FATAL: config not changed due to errors.")


def persist_results(pathContentsMap):
    for path, content in pathContentsMap.items():
        path.write_text(content)
